unit Cartography;

{
2000 May 28  First version
2000 May 29  Trap non-existant data files correctly

Copyright  David J Taylor, Edinburgh
Web site:  www.satsignal.net
E-mail:    davidtaylor@writeme.com
}

interface

uses
  Classes, Graphics;

const
  boundary_data_filename = 'Countries.dat';

const
  clBorder = clYellow;
  clLand = clGreen;
  clWater = clNavy;
  clGrid = clRed;

type
  PSmallPoint = ^TSmallPoint;  // The country boundary data is 16-bit
  TSmallPoint = record         // and here's and X, Y pair of them
    x, y: SmallInt;
  end;
  PSmallPointArray = ^TSmallPointArray;
  TSmallPointArray = array [0..0] of TSmallPoint;  // here's a bunch of 16-bit points

type
  TLineSegment = class (TObject)
    maxNlat: integer;
    maxSlat: integer;
    maxElong: integer;
    maxWlong: integer;
    count: longint;
    flags: byte;       // 76543210 - bit field split as follows
                       // 7654321  - detail levels 7..1
                       //        0 - 0 = polygons, 1 = line segments
    PPoints: PSmallPointArray;
  end;

type
  // This types extends TList to handle a list of line segments.
  // As each segment takes memory, it is important that the Clear
  // procedure (called by Destroy) releases that memory.
  TCartogList = class (TList)
  private
    FMaxPoints: integer;
    procedure ReadData (const filename: string);
  protected
    function GetLine (Index: Integer): TLineSegment;
  public
    property Lines [Index: Integer]: TLineSegment read GetLine;
    property MaxPoints: integer read FMaxPoints;
    constructor Create;
    function AddLine (ALine: TLineSegment): Integer;
    procedure Clear;  override;
    procedure LoadFromFile (const Filename: string);
    procedure LoadFromStream (Stream: TStream); virtual;
  end;


implementation

uses
  SysUtils, Forms;

{ TCartogList }

function TCartogList.AddLine (ALine: TLineSegment): Integer;
begin
  Result := Inherited Add (ALine);
end;

procedure TCartogList.Clear;
var
  p: integer;
  line: TLineSegment;
begin
  for p := 0 to Count - 1 do
    begin
    line := Items [p];
    FreeMem (line.PPoints);
    line.Free;
    end;
  FMaxPoints := 0;
  Inherited;
end;

constructor TCartogList.Create;
begin
  Inherited Create;
  FMaxPoints := 0;
end;

function TCartogList.GetLine (Index: Integer): TLineSegment;
begin
  Result := Items [Index];
end;

procedure TCartogList.LoadFromFile (const Filename: string);
begin
  Clear;
  ReadData (Filename);
end;

procedure TCartogList.LoadFromStream (Stream: TStream);
// Stores the maximum number of points in a single
// line segment, or zero if the routine fails.
var
  eofile: integer;
  done: boolean;
  lines: TLineSegment;
  line_bytes: integer;
  i: integer;
begin
  Clear;
  try
    eofile := Stream.size;
    done := False;
    while not done do
      begin
      lines := TLineSegment.Create;
      lines.maxnlat := 0;
      lines.maxslat := 0;
      lines.maxelong := 0;
      lines.maxwlong := 0;
      Stream.Read (lines.count, SizeOf (Longint));
      if lines.count > FMaxPoints then FMaxPoints := lines.count;
      Stream.Read (lines.flags, SizeOf (Byte));
      if Stream.position = eofile
      then
        begin
        Done := True;
        lines.Free;
        end
      else
        begin
        line_bytes := SizeOf (TSmallPoint) * lines.count;
        GetMem (lines.PPoints, line_bytes);
        Stream.Read (lines.PPoints^, line_bytes);
        with lines do
          begin
          for i := 0 to count - 1 do
            begin
            with PPoints^ [i] do
              begin
              if x < 0
              then
                begin
                if x < maxWlong then maxWlong := x;
                end
              else
                begin
                if x > maxElong then maxElong := x;
                end;
              if y < 0
              then
                begin
                if y < maxSlat then maxSlat := y;
                end
              else
                begin
                if y > maxNlat then maxNlat := y;
                end;
              end;
            end;
          end;
        AddLine (lines);
      end;
    end;
  except
  end;
end;

procedure TCartogList.ReadData (const filename: string);
// Read the country/river/lake data....
var
  stream: TMemoryStream;
begin
  stream := TMemoryStream.Create;
  try
    stream.LoadFromFile (filename);
    LoadFromStream (stream);
  except
  end;
  stream.Free;
end;


end.
