unit MapReaderForm;

{
Purpose:

This program reads both country boundary data from COUNTRIES.DAT and
topographic data from the US Navy 5 minute file.  The boundaries are used to
fill a bitmap with country and lake data.  This bitmap is subsequently
scanned to determine the heights for each pixel using ETOPO5 data.

For efficiency, the topography file is read using memory-mapped I/O.


The file names are now:

  countries.dat .... country and state boundaries
  etopo5.elev ...... 5' topography


Data sources:

  countries.dat .... www.satsignal.net => Satellite Tools, WXtrack =>
                       Downloads, Country & State boundary data
  etopo5.elev ...... http://www.geol.binghamton.edu/faculty/jones/etopo5.zip

  ETOPO5.elev ...... http://edcwww.cr.usgs.gov/Webglis/glisbin/guide.pl/glis/hyper/guide/etopo5
  ETOPO5 ........... http://www.geol.binghamton.edu/faculty/jones/etopo5.zip
  ETOPO5.dat ....... ftp://ftp.ngdc.noaa.gov/MGG/relief/reliefcd/SEARCH/
  TOPO_8.2.img ..... ftp://topex.ucsd.edu/pub/global_topo_2min/


Revision history:

V1.0.0  1997 Jan 05  First version
V1.0.2  1997 Feb 16  Add monochrome option
V1.0.4  1997 Jun 15  Add visible topography update by extra bitmap & CopyRect
                     Add Icon and version information
V1.0.6  1997 Jul 05  Revise icon
V2.0.0  1999 Jan 26  Version for Delphi 4
V2.0.2  1999 Jan 29  Change sea colour for WinOrbit map
V2.1.0  1999 Jun 06  First released version
                     Drop support for 10-minute maps
                     Drop support for DJT-style topography
                     Use either style of "official" ETOPO5 database
                     Look for data files in several places
V2.1.1  1999 Oct 23  Add save of basic data
                     Add state boundaries
V2.2.0  2001 Sep 17  Version for Delphi 5
                     Use current version of COUNTRIES.DAT
                     Add support for NOAA/MGG version of ETOPO5
                     Optimise display update timing when adding topography
                     Use separate Cartography unit
V2.3.0  2001 Sep 18  Add support for 2' data in file topo_8.2.img
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, Menus, ComCtrls, StdCtrls, Math,
  Cartography;

type
 TTopoStyles = (tsUnknown, tsJones, tsNGDC, tsNOAA, tsTopo2);

type
  data_source = (primary, primary_up, secondary, secondary_up, old);

const
  first_data = primary;
  last_data = old;

type
  TFormMain = class(TForm)
    Image1: TImage;
    Timer1: TTimer;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Exit1: TMenuItem;
    N1: TMenuItem;
    Save1: TMenuItem;
    ProgressBar1: TProgressBar;
    View: TMenuItem;
    Presets: TMenuItem;
    OpenWorld: TMenuItem;
    OpenMiddleEast: TMenuItem;
    OpenUK: TMenuItem;
    OpenScandinavia: TMenuItem;
    ViewTopography: TMenuItem;
    Zoomout1: TMenuItem;
    Monochrome1: TMenuItem;
    LabelDataSources: TLabel;
    procedure FormCreate (Sender: TObject);
    procedure Timer1Timer (Sender: TObject);
    procedure Image1MouseDown (Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Image1MouseUp (Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Exit1Click (Sender: TObject);
    procedure Save1Click (Sender: TObject);
    procedure OpenWorldClick (Sender: TObject);
    procedure OpenUKClick (Sender: TObject);
    procedure OpenScandinaviaClick (Sender: TObject);
    procedure OpenMiddleEastClick (Sender: TObject);
    procedure FormDestroy (Sender: TObject);
    procedure ViewTopographyClick (Sender: TObject);
    procedure Image1MouseMove (Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure Zoomout1Click (Sender: TObject);
    procedure Monochrome1Click (Sender: TObject);
  private
    { Private declarations }
    where_x, where_y: integer;
    old_x, old_y: integer;
    capturing: boolean;
    colours: array [SmallInt] of TColor;
    monochrome: boolean;

    topo_file_name: string;
    topo_file_handle: THandle;
    topo_file_mapping: THandle;
    topo_file_base: Pointer;

    topo_style: TTopoStyles;
    cartog_list: TCartogList;
    data_folder: array [first_data..last_data] of string;
    cartog_file_name: string;

    long_base: integer;           //   left of image, W..E -10800 .. +10800
    lat_base: integer;            // bottom of image, N..S  +5400 ..  -5400
    long_limit: integer;          //  right of image, W..E -10800 .. +10800
    lat_limit: integer;           //    top of image, N..S  +5400 ..  -5400

    x_minutes_per_pixel: extended;
    y_minutes_per_pixel: extended;

    procedure create_colours;
    procedure read_and_display_topo;
    function locate_topographic_data: boolean;
    function get_cartographic_data: boolean;
    procedure set_cartographic_scale;
    procedure display_cartographic_data;
    function map_long (long: integer): integer;   // Map long to pixel location
    function map_lat (lat: integer): integer;
    function map_x (x: integer): integer;         // Map pixel location to long
    function map_y (y: integer): integer;
  public
    { Public declarations }
  end;

var
  FormMain: TFormMain;

implementation

{$R *.DFM}

procedure TFormMain.FormCreate (Sender: TObject);
begin
  topo_file_handle := INVALID_HANDLE_VALUE;

  data_folder [primary] := ExtractFilePath (Application.ExeName);
  data_folder [primary_up] := data_folder [primary] + '..\';
  data_folder [secondary] := 'C:\Davids\Delphi5\Radio\WXtrack\Source\';
  data_folder [secondary_up] := data_folder [secondary] + '..\';
  data_folder [old] := 'C:\Davids\Delphi3\Graphics\WorldMap\';
  data_folder [old] := 'F:\GTopo2\';

  topo_style := tsUnknown;
  if not locate_topographic_data
    then ShowMessage ('Topography will not be available');
  get_cartographic_data;

  LabelDataSources.Hint := 'Topo:    ' + topo_file_name + #13 +
                           'Cartog: ' + cartog_file_name;

  // Initialise data to whole world mapping co-ordinates
  long_base := -180*60;     // in minutes, i.e. 180W
  lat_base := -90*60;       // in minutes, i.e. 90S
  long_limit :=  180*60;    // in minutes, i.e. 180W
  lat_limit :=  90*60;      // in minutes, i.e. 90S

  Timer1.Enabled := True;
end;


procedure TFormMain.FormDestroy (Sender: TObject);
begin
  if Assigned (cartog_list) then cartog_list.Free;
  if topo_file_handle <> INVALID_HANDLE_VALUE then
  begin
    UnmapViewOfFile (topo_file_base);
    CloseHandle (topo_file_mapping);
    CloseHandle (topo_file_handle);
  end;
end;


procedure TFormMain.Timer1Timer (Sender: TObject);
begin
  Timer1.Enabled := False;
  display_cartographic_data;
end;


function TFormMain.locate_topographic_data: boolean;
const
  num_names = 4;
  data_file_names: array [1..num_names] of string =
    ('etopo5.elev', 'etopo5.dat', 'etopo5', 'topo_8.2.img');
var
  found: boolean;
  source: data_source;
  peek1: word;
  peek2: word;
  f: file of word;
  records: integer;
  saved_filemode: integer;
  name: integer;
begin
  found := False;
  for name := 1 to num_names do
    begin
    for source := first_data to last_data do
      begin
      if found then Break;
      found := FileExists (data_folder [source] + data_file_names [name]);
      if found then
        begin
        topo_file_name := data_folder [source] + data_file_names [name];
        saved_filemode := FileMode;
        FileMode := 0;
        AssignFile (f, topo_file_name);
        Reset (f);
        records := FileSize (f);
        Read (f, peek1);
        Seek (f, 4320);
        Read (f, peek2);
        CloseFile (f);
        FileMode := saved_filemode;
        if records = 68428800 then
          topo_style := tsTopo2;
        if records = 180 * 12 * (360 * 12 + 1) then
          if peek1 = $EF3E then topo_style := tsJones;
        if records = 180 * 12 * 360 * 12 then
          begin
          if peek1 = $3EEF then
            begin
            case peek2 of
              $40EF: topo_style := tsNOAA;
              $4AEF: topo_style := tsNGDC;
            else
              topo_style := tsUnknown;
            end;
            end;
          end;
        end;
      if found then Break;
      end;
    end;
  if not found then ShowMessage ('Cannot find any topography data');
  Result := found and (topo_style <> tsUnknown);
end;


function TFormMain.get_cartographic_data: boolean;
var
  found: boolean;
  source: data_source;
begin
  cartog_file_name := boundary_data_filename;  // From cartography unit
  cartog_list := TCartogList.Create;
  for source := first_data to last_data do
    begin
    cartog_file_name := data_folder [source] + boundary_data_filename;
    found := FileExists (cartog_file_name);
    if found then
      begin
      cartog_list.LoadFromFile (cartog_file_name);
      Break;
      end;
    end;
  if not found then
    begin
    ShowMessage ('Cannot find any cartography data.  Please check the ' +
                 'location for the file ' + boundary_data_filename);
    FreeAndNil (cartog_list);
    cartog_file_name := '';
    end;
  Result := found;
end;


procedure TFormMain.create_colours;

// Procedure to create "best" colour mapping.  This is made more complicated
// by the lack of distinction between land and sea in either the high or low
// resolution files, and the presence of bathymetry data in the high-res file
// Here, we use two greens for -100' and -200' and the usual colours for
// land masses

  procedure interpolate_colours (p, q: integer);
  var
    r, g, b: integer;
    r1, g1, b1: integer;
    r2, g2, b2: integer;
    colour: integer;
  begin
    b1 := colours [p] shr 16;
    g1 := (colours [p] shr 8) and $FF;
    r1 := colours [p] and $FF;
    b2 := colours [q] shr 16;
    g2 := (colours [q] shr 8) and $FF;
    r2 := colours [q] and $FF;
    for colour := p+1 to q-1 do
      begin
      r := ((q - colour) * r1 + (colour - p) * r2) div (q - p);
      g := ((q - colour) * g1 + (colour - p) * g2) div (q - p);
      b := ((q - colour) * b1 + (colour - p) * b2) div (q - p);
      colours [colour] := b shl 16 + g shl 8 + r;
      end;
  end;

const
  min_interval = -2;
  max_interval =  8;
type
  units = (metres, feet);
  intervals = min_interval..max_interval;

const
  heights: array [units] of
    array [intervals] of integer = (
    (-200, -100, 0, 300, 600, 1000, 1500, 3000, 6000, 10000, 16000),
    (-200, -100, 0, 300, 600, 1000, 1500, 3000, 6000, 10000, 16000)
  );
  shades: array [intervals] of TColor = (
    133 shl 16 + 219 shl 8 +  55,     //  -200
     95 shl 16 + 222 shl 8 +  98,     //  -100
     57 shl 16 + 225 shl 8 + 141,     //     0
    137 shl 16 + 237 shl 8 + 187,     //   300
    194 shl 16 + 245 shl 8 + 220,     //   600
    188 shl 16 + 216 shl 8 + 224,     //  1000
    154 shl 16 + 196 shl 8 + 209,     //  1500
    104 shl 16 + 176 shl 8 + 230,     //  3000
     79 shl 16 + 150 shl 8 + 170,     //  6000
    185 shl 16 + 149 shl 8 + 182,     // 10000
    clWhite                           // 16000
  );

  procedure create_height_colours (height_units: units);
  var
    band: min_interval..max_interval;
    h_previous: integer;
    height: integer;
  begin
    // Do the stuff below sea level to the minimum colour
    for height := -32768 to 0 do
      colours [height] := shades [min_interval];
    // Sweep height pairs across remaining bands, i.e. those for which
    // contours are defined.  Define the height in the current band, and
    // set that at a preset colour, remembering that the heights are byte
    // swapped in the data file.  For all bands except the first one, perform
    // a smooth interpolation between the current height/colour and the
    // previous height/colour.  This improves the visual apperance of map
    // somewhat.
    h_previous := 0;          // just to stop the compiler warning message...
    for band := min_interval to max_interval do
      begin
      height := heights [height_units, band];
      colours [height] := shades [band];
      if band <> min_interval then
        interpolate_colours (h_previous, height);
      h_previous := height;
      end;
    // Do the stuff above the max height - set it all white
    for height := h_previous to 32767 do
      colours [height] := shades [max_interval];
  end;

var
  pixel: integer;
begin
  // Compute the metre values knowing the feet
  for pixel := min_interval to max_interval do
    heights [metres, pixel] := round (25.4 * 12 * heights [feet, pixel] / 1000);
  case topo_style of
    tsNGDC, tsNOAA, tsTopo2: create_height_colours (metres);
           tsJones: create_height_colours (metres);
  end;
end;


procedure TFormMain.read_and_display_topo;

  function MercatorMapMinutes (const iLat: integer): integer;
  const
    Deg2Rad = pi / 180.0;
    Rad2Deg = 180.0 / pi;
    pi4 = pi / 4;
  var
    lat: single;
  begin
    lat := iLat / 60.0;
    // Code to change from linear to Mercator mapping.  Note that values
    // around zero latitude could cause errors if tan (45) is just less
    // than 1!  To trap this, use a try..except, and if statements.
    try
      if lat > 0      // North of the equator
      then
        lat := Rad2Deg * ln (tan (pi4 + Deg2Rad * 0.5 * lat))
      else
        if lat < 0    // South of the equator
          then lat := -Rad2Deg * ln (tan (pi4 - Deg2Rad * 0.5 * lat));
    except
      // The value of lat will do as it is, would have been very small.
    end;
    Result := Round (lat * 60.0);
  end;

type
  TSmallIntArray = array [0..0] of SmallInt;
  PSmallIntArray = ^TSmallIntArray;
const
  clOcean: TColor = TColor ($F0F080);
var
  p: PSmallIntArray;
  line, pixel: integer;
  lon, lat: integer;
  value: TColor;
  temp_bitmap: TBitmap;
  rect: TRect;
  ticks, t0: DWORD;
begin
  if topo_style = tsUnknown then Exit;

  // Get a temporary copy of the picture of the cartographic data
  temp_bitmap := TBitmap.Create;
  temp_bitmap.Assign (Image1.Picture);   // copy the data

  // We don't know which topograhy file the user has chosen - unmap any existing
  if topo_file_handle <> INVALID_HANDLE_VALUE then
  begin
    UnmapViewOfFile (topo_file_base);
    CloseHandle (topo_file_mapping);
    CloseHandle (topo_file_handle);
  end;

  // Create a mapped file of the topography data
  topo_file_handle := CreateFile (PChar (topo_file_name),
                              GENERIC_READ, FILE_SHARE_READ, nil,
                              OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

  if topo_file_handle = INVALID_HANDLE_VALUE then Exit;

  topo_file_mapping := CreateFileMapping (topo_file_handle, nil, PAGE_READONLY, 0, 0, nil);
  if topo_file_mapping = 0 then Exit;

  topo_file_base := MapViewOfFile (topo_file_mapping, FILE_MAP_READ, 0, 0, 0);
  if topo_file_base = nil then Exit;

  create_colours;

  // Set the limits and position of the progress bar - hardly required now....
  with ProgressBar1 do
    begin
    Min := 0;
    Max := Image1.Height - 1;
    Position := Min;
    Visible := True;
    end;

  // This rectangle defines the area to be updated on the main image
  rect.Left := 0;
  rect.Right := Image1.Width;
  rect.Top := 0;
  rect.Bottom := 0;

  t0 := GetTickCount;
  for line := 0 to Image1.Height - 1 do
    begin
    Inc (rect.Bottom);            // Another line, update the region definition
    p := topo_file_base;          // Point to start of all topographic data
    lat := map_y (line);          // Convert line number to latitude in minutes
    if (lat > -5400) and (lat <= 5400) then
    case topo_style of
      tsTopo2:
        begin
        lat := MercatorMapMinutes (lat);
        lat := (6336 - lat) div 2;
        if (lat >= 0) and (lat < 6336)
          then Inc (p, lat * 10800)   // Start of active data for this latitude
          else p := nil;
        end;
      tsNGDC, tsNOAA:
        begin
        lat := (5400 - lat) div 5;
        Inc (p, lat * 4320);   // Start of active data for this latitude
        end;
      tsJones:
        begin
        lat := (5400 - lat) div 5;
        Inc (p, lat * (4320 + 1));   // Start of active data for this latitude
        end;
    end;

    // Convert all pixels in the line to their height colour
    if Assigned (p) then
      for pixel := 0 to Image1.width - 1 do
      begin
      value := temp_bitmap.Canvas.Pixels [pixel, line];
      case value of
        clWhite, clAqua, clBlue:
          temp_bitmap.Canvas.Pixels [pixel, line] := clOcean;
        else
          begin
          lon := map_x (pixel);
          if (lon > -10800) and (lon <= 10800) then
            case topo_style of
              tsJones:
                begin
                lon := ((lon + 21600) div 5) mod 4320;
                temp_bitmap.Canvas.Pixels [pixel, line] := colours [p^ [lon]];
                end;
              tsNGDC:
                begin
                lon := (lon + 10800) div 5;
                temp_bitmap.Canvas.Pixels [pixel, line] := colours [Swap (p^ [lon])];
                end;
              tsNOAA:
                begin
                lon := ((lon + 21600) div 5) mod 4320;
                temp_bitmap.Canvas.Pixels [pixel, line] := colours [Swap (p^ [lon])];
                end;
              tsTopo2:
                begin
                lon := ((lon + 21600) div 2) mod 10800;
                temp_bitmap.Canvas.Pixels [pixel, line] := colours [Swap (p^ [lon])];
                end;
            end;
          end
      end;
      end;

    // Every so often, update the main image, do this in blocks to
    // minimise the amount of bit manipulation that must go on.....
    ticks := GetTickCount;
    if (ticks - t0) > 200 then
      begin
      t0 := ticks;
      ProgressBar1.Position := line;
      ProgressBar1.Update;
      Image1.Canvas.CopyRect (rect, temp_bitmap.Canvas, rect);  // copy a block
      Image1.Update;                                            // and show it
      rect.Top := rect.Bottom;
      end;
    end;

  // A final update in case the last block wasn't done
  Image1.Canvas.CopyRect (rect, temp_bitmap.Canvas, rect);
  Image1.Update;
  temp_bitmap.Free;
  ProgressBar1.Visible := False;
end;


procedure TFormMain.Exit1Click (Sender: TObject);
begin
  Close;
end;


procedure TFormMain.Save1Click (Sender: TObject);
begin
  Image1.Picture.SaveToFile ('temp.bmp');
end;


function TFormMain.map_long (long: integer): integer;
begin
  Result := Round ((long - long_base) / x_minutes_per_pixel);
end;


function TFormMain.map_x (x: integer): integer;
begin
  Result := Round (x * x_minutes_per_pixel + long_base);
end;


function TFormMain.map_lat (lat: integer): integer;
begin
  Result := Image1.Height - Round ((lat - lat_base) / y_minutes_per_pixel);
end;


function TFormMain.map_y (y: integer): integer;
begin
  Result := Round ((Image1.Height - y) * y_minutes_per_pixel + lat_base);
end;


procedure TFormMain.set_cartographic_scale;
var
  v, h: extended;
  c: extended;
  i: integer;
begin
  // Ensure the rectangle is specified (left, bottom) .. (right, top)
  if lat_base > lat_limit then
    begin
    i := lat_base;
    lat_base := lat_limit;
    lat_limit := i;
    end;

  if long_base > long_limit then
    begin
    i := long_base;
    long_base := long_limit;
    long_limit := i;
    end;

  if long_base < -10800 then long_base := -10800;
  if long_limit < -10800 then long_limit := -10800;
  if long_base > 10800 then long_base := 10800;
  if long_limit > 10800 then long_limit := 10800;

  if lat_base < -5400 then lat_base := -5400;
  if lat_limit < -5400 then lat_limit := -5400;
  if lat_base > 5400 then lat_base := 5400;
  if lat_limit > 5400 then lat_limit := 5400;

  // Get mean viewing angle
  c := (lat_base + lat_limit)/ 2;
  // and the vertical multiplying factor to use
  c := 1.0 / cos (1.5708 * c / 5400.0);

  // Now compute the suggested minutes per pixel in each direction
  h := (long_limit - long_base) / Image1.Width;
  v := c * (lat_limit - lat_base) / Image1.Height;

  if v > h
    then x_minutes_per_pixel := v
    else x_minutes_per_pixel := h;
  y_minutes_per_pixel := x_minutes_per_pixel / c;
end;


procedure TFormMain.display_cartographic_data;

// In this data, the polygons and polylines are in 16-bit
// minutes, ranging from -10800 (180 W) ... 0 (0) ... +10800 (180 E)
//              and from  +5400  (90 N) ... 0 (0) ...  -5400  (90 S)

type
  PPointArray = ^TPointArray;
  TPointArray = array [0..0] of TPoint;
var
  pVertices: PPointArray;
  poly: integer;
  point: integer;
  poly_type: integer;
begin
  set_cartographic_scale;
  if not Assigned (cartog_list) then Exit;
  if cartog_list.Count = 0 then Exit;

  GetMem (pVertices, cartog_list.MaxPoints * SizeOf (TPoint));
  Image1.Visible := False;

  with cartog_list, Image1.Canvas do
    begin
    if monochrome
      then Brush.Color := clWhite
      else Brush.Color := clBlue;
    Pen.Color := Brush.Color;
    Brush.Style := bsSolid;
    Pen.Style := psSolid;
    Pen.Color := clBlack;
    Rectangle (0, 0, Image1.Width, Image1.Height);
    for poly := 0 to Count - 1 do
      begin
      with cartog_list.Lines [poly] do
        begin
        for point := 0 to count - 1 do
          begin
          with PPoints^ [point] do
            begin
            pVertices^ [point].x := map_long (x);
            pVertices^ [point].y := map_lat (y);
            end;
          end;
//        Brush.Style := bStyle;
//        Pen.Color := pColour;
//        Pen.Style := pStyle;
        poly_type := flags and $FE;

        // Do countries (0) and lakes (8)
        // 4 - state boundaries
        // 16 - rivers
        if monochrome
        then
          case poly_type of
            0: Brush.Color := RGB (192, 192, 192);
            8: Brush.Color := clWhite;
          end
        else
          case poly_type of
            0: Brush.Color := RGB (0, 128, 0);
            8: Brush.Color := RGB (0, 255, 255);
          end;

        case poly_type of
          0, 4, 8: if (flags and 1) = 1
             then Polyline (Slice (pVertices^, count))
             else Polygon (Slice (pVertices^, count));
        end;

        end;
      end;
    end;

  FreeMem (pVertices);
  Image1.Visible := True;
end;


procedure TFormMain.OpenWorldClick (Sender: TObject);
begin
  long_base := -180*60;             // in minutes, i.e. 180W
  lat_base := -90*60;               // in minutes, i.e. 90S
  long_limit :=  180*60;            // in minutes, i.e. 180W
  lat_limit :=  90*60;              // in minutes, i.e. 90S
  display_cartographic_data;
end;


procedure TFormMain.OpenUKClick (Sender: TObject);
begin
  long_base := -11 * 60;            // in minutes, i.e. 11 W
  lat_base := round (49.5 * 60);    // in minutes, i.e. 49.5 N
  long_limit := 2 * 60;             // in minutes, i.e. 2 W
  lat_limit := 59 * 60;             // in minutes, i.e. 59 N
  display_cartographic_data;
end;


procedure TFormMain.OpenScandinaviaClick (Sender: TObject);
begin
  long_base := -2 * 60;             // in minutes, i.e. 11W
  lat_base := round (53.5 * 60);    // in minutes, i.e. 53.5N
  long_limit := +13 * 60;
  lat_limit := round (70.5 * 60);
  display_cartographic_data;
end;


procedure TFormMain.OpenMiddleEastClick (Sender: TObject);
begin
  long_base := 16 * 60;             // in minutes, i.e. 16E
  long_limit := 110 * 60;
  lat_base := 4 * 60;               // in minutes, i.e. 4N
  lat_limit := 20 * 60;
  display_cartographic_data;
end;


procedure TFormMain.ViewTopographyClick (Sender: TObject);
begin
  display_cartographic_data;
  read_and_display_topo;
end;


procedure TFormMain.Image1MouseDown (Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  where_x := X;
  where_y := Y;
  old_x := X;
  old_y := Y;
  SetCaptureControl (Image1);
  with Image1.Canvas, Image1.Canvas.Pen do
    begin
    Color := clAqua;
    Mode := pmXor;
    Style := psDash;
    Rectangle (where_x, where_y, old_x, old_y);
    end;
  capturing := True;
end;


procedure TFormMain.Image1MouseMove (Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if capturing then with Image1.Canvas do
    begin
    Rectangle (where_x, where_y, old_x, old_y);   // undraw the old rectangle
    old_x := X;
    old_y := Y;
    Rectangle (where_x, where_y, old_x, old_y);   // draw the new one
    end;
end;


procedure TFormMain.Image1MouseUp (Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  long0, lat0: integer;
  long1, lat1: integer;
begin
  if capturing then with Image1.Canvas, Image1.Canvas.Pen do
    begin
    Rectangle (where_x, where_y, old_x, old_y);   // undraw the old rectangle
    Mode := pmCopy;
    Style := psSolid;
    end;
  capturing := False;
  SetCaptureControl (nil);
  long0 := map_x (where_x);
  lat0 := map_y (where_y);
  if old_x < 0 then old_x := 0;
  if old_x > Image1.Width then old_x := Image1.Width;
  if old_y < 0 then old_y := 0;
  if old_y > Image1.Height then old_y := Image1.Height;
  long1 := map_x (old_x);
  lat1 := map_y (old_y);
  lat_base := lat0;
  lat_limit := lat1;
  long_base := long0;
  long_limit := long1;
  Timer1.Enabled := True;
end;


procedure TFormMain.Zoomout1Click (Sender: TObject);
var
  long_range, lat_range: integer;
begin
  long_range := long_limit - long_base;
  long_base := long_base - long_range div 2;
  long_limit := long_limit + long_range div 2;
  lat_range := lat_limit - lat_base;
  lat_base := lat_base - lat_range div 2;
  lat_limit := lat_limit + lat_range div 2;
  Timer1.Enabled := True;
end;


procedure TFormMain.Monochrome1Click (Sender: TObject);
begin
  Monochrome1.Checked := not Monochrome1.checked;
  monochrome := Monochrome1.Checked;
  display_cartographic_data;
end;

end.

