unit TileData;

{
Revision history:

2000 Apr 08  First version

Copyright  David J Taylor, Edinburgh
Web site:  http://www.davidtaylor.freeserve.co.uk
E-mail:    davidtaylor@writeme.com
}

interface

uses
  Windows, Classes, SysUtils;

const
  rows = 6000;
  cols = 4800;
  array_rows = 180 * 60 * 2;
  array_cols = 360 * 60 * 2;

type
  TElevation = SmallInt;

type
  TTile = class
  private
    FFileHandle: THandle;
    FFileMapping: THandle;
    FBaseAddress: Pointer;
    FColumns: integer;
    FBytesPerScanline: integer;
    function GetScanline (Row: Integer): Pointer;
  public
    constructor Create (const Filename: string);
    destructor Destroy; override;
    property ScanLine [Row: Integer]: Pointer read GetScanLine;
  end;

  TTileArray = class
  private
    FBaseFolder: string;
    FTilesN90: array [0..8] of TTile;
    FTilesN40: array [0..8] of TTile;
    FTilesS10: array [0..8] of TTile;
    FTilesS60: array [0..5] of TTile;
  public
    constructor Create (const BaseFolder: string);
    destructor Destroy; override;
    function HeightAt (const x, y: integer): integer;
  end;


implementation

uses
  Dialogs;

{ TTile }

constructor TTile.Create (const Filename: string);
begin
  Inherited Create;
  // Create a mapped file of the elevation data
  FFileHandle := CreateFile (PChar (Filename),
                             GENERIC_READ, FILE_SHARE_READ, nil,
                             OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);

  // Check we got the connection OK
  if FFileHandle = INVALID_HANDLE_VALUE
    then Raise EFOpenError.Create ('Cannot open file: ' + Filename);

  FFileMapping := CreateFileMapping (FFileHandle, nil, PAGE_READONLY, 0, 0, nil);
  if FFileMapping = 0 then Fail;

  FBaseAddress := MapViewOfFile (FFileMapping, FILE_MAP_READ, 0, 0, 0);
  if FBaseAddress = nil then Fail;

  if Pos ('0S60', UpperCase (Filename)) <> 0
    then FColumns := 7200
    else FColumns := 4800;

  FBytesPerScanline := SizeOf (SmallInt) * FColumns;
end;


destructor TTile.Destroy;
begin
  if FFileHandle <> INVALID_HANDLE_VALUE then
  begin
    UnmapViewOfFile (FBaseAddress);
    CloseHandle (FFileMapping);
    CloseHandle (FFileHandle);
  end;

  Inherited;
end;


function TTile.GetScanline (Row: Integer): Pointer;
var
  ppix: ^SmallInt;
  pix1, pix2: SmallInt;
begin
  Integer (Result) := Integer (FBaseAddress) + Row * FBytesPerScanline;
  // Access pixels at each end of the line to
  // ensure they are paged into memory.
  ppix := Result;
  pix1 := Swap (ppix^);
  Inc (ppix, FColumns - 1);
  pix2 := Swap (ppix^);
  if (pix1 < -9999) or (pix2 < -9999) then ShowMessage ('Bad Data');
end;


{ TTileArray }

constructor TTileArray.Create (const BaseFolder: string);
var
  lon: integer;
  name: string;
  filename: string;
begin
  Inherited Create;
  FBaseFolder := ExtractFilePath (BaseFolder);
  for lon := 0 to 8 do
    begin
    case lon of
      0: name := 'W180';
      1: name := 'W140';
      2: name := 'W100';
      3: name := 'W060';
      4: name := 'W020';
      5: name := 'E020';
      6: name := 'E060';
      7: name := 'E100';
      8: name := 'E140';
    end;
    try
      filename := FBaseFolder + name + 'N90.dem';
      FTilesN90 [lon] := TTile.Create (filename);
    except
      Raise EFOpenError.Create ('Cannot open file: ' + Filename);
    end;
    try
      filename := FBaseFolder + name + 'N40.dem';
      FTilesN40 [lon] := TTile.Create (filename);
    except
      Raise EFOpenError.Create ('Cannot open file: ' + Filename);
    end;
    try
      filename := FBaseFolder + name + 'S10.dem';
      FTilesS10 [lon] := TTile.Create (filename);
    except
      Raise EFOpenError.Create ('Cannot open file: ' + Filename);
    end;
    end;
  for lon := 0 to 5 do
    begin
    case lon of
      0: name := 'W180';
      1: name := 'W120';
      2: name := 'W060';
      3: name := 'W000';
      4: name := 'E060';
      5: name := 'E120';
    end;
    try
      filename := FBaseFolder + name + 'S60.dem';
      FTilesS60 [lon] := TTile.Create (filename);
    except
      Raise EFOpenError.Create ('Cannot open file: ' + Filename);
    end;
    end;
end;


destructor TTileArray.Destroy;
var
  lon: integer;
begin
  for lon := 0 to 8 do
    begin
    if Assigned (FTilesN90 [lon]) then FTilesN90 [lon].Free;
    if Assigned (FTilesN40 [lon]) then FTilesN40 [lon].Free;
    if Assigned (FTilesS10 [lon]) then FTilesS10 [lon].Free;
    end;
  for lon := 0 to 5 do
    if Assigned (FTilesS60 [lon]) then FTilesS60 [lon].Free;
  Inherited;
end;


function TTileArray.HeightAt (const x, y: integer): integer;
var
  tile_row, tile_col: integer;
  row: ^SmallInt;
begin
  tile_row := y div 6000;
  if tile_row = 3
    then tile_col := x div 7200
    else tile_col := x div 4800;
  case tile_row of
    0: begin
       row := FTilesN90 [tile_col].ScanLine [y mod 6000];
       Inc (row, x mod 4800);
       Result := Swap (row^);
       end;
    1: begin
       row := FTilesN40 [tile_col].ScanLine [y mod 6000];
       Inc (row, x mod 4800);
       Result := Swap (row^);
       end;
    2: begin
       row := FTilesS10 [tile_col].ScanLine [y mod 6000];
       Inc (row, x mod 4800);
       Result := Swap (row^);
       end;
    3: begin
       row := FTilesS60 [tile_col].ScanLine [y mod 6000];
       Inc (row, x mod 7200);
       Result := Swap (row^);
       end;
  end;
end;

end.

